\ TASKER1.F      Simple multitasking for Win32Forth

( eForth Multitask on Win32Forth )

\ 960903 tjz minor changes and adjustments for Win32Forth
\        renamed HAT to NEWTASK for readability
\ 960828 bee adapt to w32f33a7
\ 960811 bee adapt to w32f32
\        BRANCH is relative
\        note use of ABS>REL and REL>ABS
\        lomem \user\ --<sp --<rp himem
\        changed BUILD to ALSOTASK to avoid name conflict
\ 0 value mm0
0 value mm1
0 value mm2
0 value mm3
needs user_v.f
ONLY FORTH ALSO DEFINITIONS

DECIMAL
also user-lib
\ create vvv  200 allot
defer-us big-pause
defer-us end-pause
variable sp0_
variable TOR
previous

: w-pause ekey? drop ( ekey? if ekey 'C' +k_control = abort" ^C" then) ;

' w-pause to big-pause
' noop    to end-pause

\ =============================================================

.( Multitask )

: LOCAL         ( tid a -- a ) ( index another task's local variable )
                op> RP0 - ( address to offset )
                + >op ; ( offset to address )

: LOCAL>        ( tid -- a )
           state @ if  compile  FOLLOWER
                       compile  LOCAL
                      [compile] &of
                       compile  @
                       compile  +
                   else FOLLOWER
                        LOCAL
                      [compile]  &of
                         @ +
                   then  ; immediate

INTERNAL

DODEFER CONSTANT PASS

: (WAKE)  UP! TOS @ SP! ;

CFA-FUNC WAKE         ( -- ) ( restore follower )
             (WAKE) RP! end-pause ;

EXTERNAL

: big-pause_ \ cr ." big-pause_"
             sp@ sp0 @ umin sp0 @ over -  >r sp0_ @
             r@ - r@ 0max cmove
             sp0_ @ r>  -  sp!
             sp0_ @ sp0 ! ;

: end-pause_
             sp@ sp0 @ over - >r
             CONUSER TOS local @ -4 cells+ dup  sp0 !
             r@ - r@  0max cmove
             sp0 @ r> - sp!
             \ cr ." end-pause_" xxx
               ;

: (PAUSE)         ( -- ) ( allow another task to execute )
                big-pause RP@ cell- \ TOR !
                 SP@ TOS !  FOLLOWER @ execute
               ( end-pause ) ;
: PAUSE   \ for between task goto
  noop    \ 
 ;

: STOP          ( -- ) ( sleep current task )
                PASS STATUS ! PAUSE ;

: SLEEP         ( tid -- ) ( sleep another task )
                PASS SWAP STATUS LOCAL ! ;

: AWAKE         ( tid -- ) ( wake another task )
                WAKE SWAP STATUS LOCAL ! ;

: ACTIVATE          ( tid -- )
                DUP 2@                ( tid sp rp )
                R> OVER !             ( save entry at rp )
                OVER cell- ! cell-    ( save rp at sp )
                OVER TOS LOCAL !      ( save sp in tos )
                AWAKE ;

\ =============================================================

\ Define a new task 'name', with u user area bytes, s stack byte
\ space, and r return stack space.

: NEWTASK_       ( u s r  -- ) ( -- tid )
                HERE >R  ( rp0 ) 0 , ( sp0 ) 0 ,
                ROT USERMIN + ( user area )
                + ALLOT ( return stack )
                HERE  -1 DUP , DUP , DUP , , ( tracer )
                SWAP ALLOT ( data stack )  HERE  -1 DUP , , ( tracer )
                swap R@ 2!
                up@ r> FOLLOWER LOCAL  up!
                sp0 @ sp0_ !
                ['] big-pause_ to big-pause
                ['] end-pause_ to end-pause
                decimal
                up!
                 ;

: NEWTASK       ( u s r 'name' -- ) ( -- tid )
        CREATE NEWTASK_ ;

\ initialize task tid, and put it to sleep

: ALSOTASK      ( tid -- )
                DUP CONUSER = IF DROP EXIT THEN       ( not main task )
                DUP SLEEP                             ( sleep new task )
                FOLLOWER @ OVER FOLLOWER LOCAL !      ( link new task )
                STATUS LOCAL FOLLOWER ! ;             ( link old task )

: ONLYTASK      ( -- ) ( initialize main task )
                CONUSER  DUP [ FOLLOWER RP0 - ] LITERAL + UP!
                STATUS FOLLOWER !  AWAKE ;

ONLYTASK ( initialize main task )

\ =============================================================
\ Shared resource locking

: SEMAPHORE     ( 'name' -- )   \ define a semaphore for resource arbitation
                CREATE 0 , ;

\ get a semaphore, locking out other tasks from using the resource
\ related to this semaphore

: SEM-GET       ( semaphore -- )
                PAUSE ( remember your manners )
                DUP @ STATUS XOR ( owner ? )
                IF BEGIN DUP @ WHILE PAUSE REPEAT ( no, wait for release )
                  STATUS SWAP ! ( lock ) EXIT
                THEN DROP ;

\ release a semaphore, allowing other tasks to access the resource
\ related to this semaphore

: SEM-RELEASE   ( semaphore -- )
                DUP @ STATUS XOR IF DROP EXIT THEN 0 SWAP ! ( unlock ) ;
2variable pauk-xy


INTERNAL

: MULTI-KEY     ( -- c )        \ new version of key for multi-tasking
                getxy pauk-xy 2!
                BEGIN PAUSE KEY? UNTIL MENUKEY
                pauk-xy 2@ gotoxy  ;

EXTERNAL

: MULTI         ( -- )          \ enable multi-tasking
                ['] MULTI-KEY IS KEY ;

: SINGLE        ( -- )          \ disable multi-tasking
                ['] MENUKEY IS KEY ;

MODULE
\S =============================================================
\  Test and Example Code Follows
\  =============================================================

#user 40  CELLS 80 CELLS NEWTASK T1
  T1 ALSOTASK

0x17c 40 CELLS 80 CELLS NEWTASK T2
\ .( zzz ) vvv @ @ >name .id
  T2 ALSOTASK

VARIABLE V1

: T1GO ( -- )
  0 V1 !
  T1 ACTIVATE
  98765
  BEGIN PAUSE 1 V1 +!
  AGAIN ;

VARIABLE V2
create  www 200 allot
: xxx ;
: T2GO ( -- )
  0 V2 !
  T2 ACTIVATE \ end-pause
  BEGIN xxx 0 0  .s ." sssssss"
    PAUSE  . .
  AGAIN ;
/*
also hidden also

: ttttt    0x34c04 cr ." tt" @ h.
 [  ' do-watch 9 cells+ @ dup >name .id , ] 0x34c04  @ h. ;

' ttttt ' do-watch 9 cells+ !
previous previous
*/

' (pause) ' pause cell+ !
 debug pause
 T1GO T2GO
: O ( -- ) V1 ? V2 ? ;
: P ( -- ) PAUSE O ;

\ =============================================================

